/*  plumb - the pipe shell - lexycal
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

%option reentrant
%option bison-bridge
%{
#define IN_LEX
#include <ctype.h>

#define YYSTYPE plumb_STYPE

#include "parse_common.h"
#include "process.h"
#include "plumb.h"
#include "var.h"
#include "config.h"



static int plumb_yy_input(parser_t *parser, char *buf, yy_size_t max_size, void *scanner);
#define YY_INPUT(buf,result,max_size) \
	do { \
		result = plumb_yy_input(yyextra.parser, buf, max_size, yyg); \
		if (result < 0) break; \
	} while(0)

typedef struct stack_s {
	YY_BUFFER_STATE yybuff;
} pb_stack_t;

typedef struct lex_parser_s {
	pb_stack_t stack[LEX_MAX_DEPTH];
	int level;
	parser_t *parser;
} lex_parser_t;

#define YY_EXTRA_TYPE lex_parser_t

static int level = 0;

void lex_push();
int lex_pop();
void lex_include();

%}

idchars [A-Za-z0-9_./-]
dquote (\\\"|[^\"])
squote (\\\'|[^\'])

%%
[#].*[\r\n]     { parser->loc.first_line++; parser->loc.first_column = 0;}
["]{dquote}*["] { yylval->str = str_quote(yytext, '\"'); parser->loc.first_column += strlen(yylval->str); return STRING; }
[']{squote}*['] { yylval->str = str_quote(yytext, '\''); parser->loc.first_column += strlen(yylval->str); return STRING; }
lib-include[ ]{idchars}+[\n]  { lex_include(yyg, parser, strclone(yytext+12), 1); }
include[ ]{idchars}+[\n]      { lex_include(yyg, parser, strclone(yytext+8), 0); }
[\\]            { }
[\\][\n]        { }
[\n]            { parser->loc.first_line++; parser->loc.first_column = 0; return NL;}
[ \t]           { parser->loc.first_column++;}
[;\r]           { parser->loc.first_column++; return ';'; }
[{}\][:,=|*+]   { parser->loc.first_column++; return *yytext; }
[0-9]+          { parser->loc.first_column += strlen(yytext); yylval->str = strclone(yytext); return INT; }
[$]{idchars}+   { parser->loc.first_column += strlen(yytext); lex_push(yyg, NULL); yy_var_subst(parser->tx, yytext, 0, yyg); }
[$][(][^\n)]*[)] { parser->loc.first_column += strlen(yytext); lex_push(yyg, NULL); yy_var_subst(parser->tx, yytext, 1, yyg); }
[$][{][^\n}]*[}] {
		char *name;
		const char *v;
		int l = strlen(yytext);
		parser->loc.first_column += l;
		name = yytext+2;
		name[l-3] = '\0';
		v = var_get(parser->tx, name);
		if (v == NULL)
			plumb_error_("Can't find variable %s\n", name);
		yylval->str = strclone(v);
		return STRING;
	}

start           { parser->loc.first_column += strlen(yytext); return KW_START; }
stop            { parser->loc.first_column += strlen(yytext); return KW_STOP; }
kill            { parser->loc.first_column += strlen(yytext); return KW_KILL; }
pause           { parser->loc.first_column += strlen(yytext); return KW_PAUSE; }
resume          { parser->loc.first_column += strlen(yytext); return KW_RESUME; }
flush           { parser->loc.first_column += strlen(yytext); return KW_FLUSH; }
exit            { parser->loc.first_column += strlen(yytext); return KW_EXIT; }
query           { parser->loc.first_column += strlen(yytext); return KW_QUERY; }
dump_internals  { parser->loc.first_column += strlen(yytext); return KW_DUMP_INTERNALS; }

{idchars}+      { parser->loc.first_column += strlen(yytext); yylval->str = strclone(yytext); return STRING; }

.               { plumb_error_("Unhandled character in input: %c (%d) - did you mean \"%c\"?\n", *yytext, *yytext, *yytext); }

<<EOF>>         { if (lex_pop(yyscanner, yyg)) { yyterminate(); } }
%%

void lex_push(struct yyguts_t *yyg)
{
	yyextra.stack[level].yybuff = YY_CURRENT_BUFFER;
	yyextra.level++;
}

int lex_pop(yyscan_t *scanner, struct yyguts_t *yyg)
{
	if (yyextra.level > 0) {
		yyextra.level--;
		yy_switch_to_buffer(yyextra.stack[level].yybuff, scanner);
		return 0;
	}
	return 1;
}

char *lex_getpos(void *scanner)
{
	struct yyguts_t *yyg = scanner;
	return yytext;
}


int lex_is_token_singlechar(int token)
{
	switch(token) {
		case NL:
			return 1;
		case INT:
		case STRING:
		case KW_STOP:
		case KW_KILL:
		case KW_PAUSE:
		case KW_RESUME:
		case KW_FLUSH:
		case KW_EXIT:
		case KW_QUERY:
		case KW_DUMP_INTERNALS:
			return 0;
		default:
			if ((token > 0) && (token < 255))
				return 1;
		plumb_error_("Internal error: lex_is_token_singlechar(): unhandled token %d\n", token);
	}
	return 0;
}

void lex_include(struct yyguts_t *yyg, parser_t *p, char *fn_, int tx_rename)
{
	transaction_t *tx = p->tx;
	char *end, *fn;

	fn = fn_;
	while(isspace(*fn)) fn++;
	end = strpbrk(fn, "\r\n");
	if (end != NULL)
		*end = '\0';

	lex_push(yyg);
	if (tx_rename) {
		int lfn, ltx;
		char *newname;

		lfn = strlen(fn);
		ltx = strlen(tx->name);
		newname = malloc(lfn + ltx + 128);
		sprintf(newname, "%s (included from %s:%ld)", fn, tx->name, p->loc.first_line+1);
		tx = tx_this_or_new(NULL, newname, tx->query_out);
		exe_append_subtx(p, tx);
	}
	yy_include(tx, NULL, fn);
	p->loc.first_line++;
	p->loc.first_column = 0;
	free(fn_);
}


/* an alternative implementation of YY_INPUT so that it doesn't need to read from a FILE * */
static int plumb_yy_input(parser_t *parser, char *buf, yy_size_t max_size, void *scanner)
{
	yyscan_t yyscanner = (yyscan_t)scanner;
	struct yyguts_t *yyg = (struct yyguts_t*)scanner;
	int result;

	if (YY_CURRENT_BUFFER_LVALUE->yy_is_interactive) {
		YY_FATAL_ERROR("plumb_yy_input: interactive mode not implemented.\n");
		abort();
	}

	errno=0;
	while((result = fread(buf, 1, max_size, yyin)) == 0 && ferror(yyin)) {
		if(errno != EINTR) {
				YY_FATAL_ERROR("plumb_yy_input: input in flex scanner failed");
				return -1;
		}
		errno=0;
		clearerr(yyin);
	}
	return result;
}

void yyset_parser_backref(parser_t *parser, void *scanner)
{
	struct yyguts_t *yyg = (struct yyguts_t*)scanner;
	yyextra.parser = parser;
}
