#ifndef PROCESS_H
#define PROCESS_H

/*************************************************/
/* process administration (both virtual and real */
/*************************************************/
#include <libporty/net/network.h>
#include "kv.h"

/* virtual and real process descriptors */
typedef struct proc_virt_s proc_virt_t;
typedef struct proc_real_s proc_real_t;
typedef struct fd_s  fd_t;
typedef struct pipe_s pipe_t;

#include "parse_common.h"

/* pipe connecting exactly two fds */
struct pipe_s {
	fd_t *from;
	fd_t *to;
	pipe_t *next, *prev;
};

typedef struct read_buf_s read_buf_t;

struct read_buf_s {
	int len;
	read_buf_t *next;
	char data[1];          /* actually len bytes long; [] is not supported by c89 */
}; 

/* an fd is a proc:local_fd pair, process may be virtual or real */
struct fd_s {
	union {
		proc_virt_t *v;
		proc_real_t *r;
	} proc;
	int fd;       /* local to the process */
	pipe_t *pipe; /* an fd is part of exactly one pipe */
	int pi;       /* poll index (in the pollfd array) */

	/* output buffering - only for sinks to hold data while waiting for POLLOUT */
	char *obuff;
	int obuff_len;
	read_buf_t *rb; /* a linked list of reads (happens on HUP) if target is blocked */
	char close_after_write;
	char to_delete;   /* delete fd from watch list after current iteration */
	char blocked;     /* 1 if can not read/write due to blocking */
	char eof;         /* whether eof is detected/sent/generated on the fd */
	char hupd;        /* source fd: already got a HUP (reading), trying to flush read buffer */
	char paused;      /* 1 if can not read/write due to pausing */
	char closing;     /* wether we are already closing the fd to avoid infinite loops */
	char vp_last_blstate;  /* last blocking state sent out to the correspoding vp (for source fds) - keep track on this so only changes are sent out */

	void (*override_do_write)(P_net_socket sockfd, fd_t *fd); /* if not NULL, call this instead of do_write() */
};

typedef enum proc_state_e {
	PS_STOPPED = 0,
	PS_RUNNING,
	PS_PAUSED
} proc_state_t;

/* common process header - both proc_real_t and proc_virt_t starts with this */
typedef struct proc_header_s {
	int virt;
	char *id;
	char *tx_name;   /* a copy of transaction name for the transaction that created the process */
	proc_state_t state;
	int num_fds, alloced_fds;
	fd_t *fds;       /* local end of our connections - remote end can be found in .pipe */

	/* property bits -  can not be bitfield because we'll need pointers to them eventually */
	char sticky;         /* plumb's exit depends on this process */
	char retval, retval_zero, retval_non_zero; /* when to set plumb's return value: always, when process returned 0 or when process returned non-0 */
	char kill_on_eof;    /* should manually kill the process after all sinks got eof */
	char respawn;        /* respawn the process when dies */

} proc_header_t;

struct proc_real_s {
	proc_header_t h;

	char **argv;          /* libporty-compatible NULL terminated command line */
	P_net_socket sock_ctrl;
	P_net_socket *sock;   /* stdio/stderr are no exception! */

	/* in extreme cases an RP may have hooks; for example _env is an RP
	   without a real UNIX process behind it */
	void (*stop)(proc_real_t *pr, int hard);

	char accept_fd;      /* accept new fd binding on the fly - for _env */

	proc_virt_t *shadow_of; /* for VPs that create RP shadows for handling real FDs */
};


typedef enum virt_ctrl_e {
	VC_OK = 0,
	VC_PAUSE,        /* virtual process wants to get paused */
	VC_QUIT          /* virtual process wants to quit */
} virt_ctrl_t;

typedef int (*vp_init_cb)(proc_virt_t *p, kvlist_t *args);

struct proc_virt_s {
	proc_header_t h;

	/* original command name that created the VP */
	char *cmd;

	/* called when input is available on one of the sinks */
	virt_ctrl_t (*input)(proc_virt_t *proc, int local_fd, const char *data, int len);

	/* called when eof detected on a sink or source */
	virt_ctrl_t (*sink_eof)(proc_virt_t *proc, int local_fd);
	virt_ctrl_t (*source_eof)(proc_virt_t *proc, int local_fd);

	/* called when a source of the current process gets blocked or the block is removed (sink is writable again) */
	virt_ctrl_t (*source_block)(proc_virt_t *proc, int local_fd, int is_blocking);

	/* called when a timer of the process triggers */
	virt_ctrl_t (*timer)(proc_virt_t *proc, void *data);

	/* called when a new pipe is connected to the process or an existing pipe is killed */
	virt_ctrl_t (*newfd)(proc_virt_t *proc, int local_fd);
	virt_ctrl_t (*delfd)(proc_virt_t *proc, int local_fd);

	/* called to free internals of this struct */
	void (*free)(proc_virt_t *porc);

	/* start the process (called after initialization, when all fd bindings are known) */
	virt_ctrl_t (*start)(proc_virt_t *proc);
	/* stop the process (called before closing all fds); does not return anything, because the process will be stopped anyway */
	void (*stop)(proc_virt_t *proc);

	transaction_t *tx;

	char stdfds[6];    /* behavior of standard fds, when enforced with vp_stdfds_set */
	void *data;
	int num_shadows;   /* number of shadow processes created from this vp - for generating unique process names for shadows */
};

/* create a new process, return pointer to the process; id must be strclone()'d */
proc_virt_t *proc_create(transaction_t *tx, char *id, int virt, char *cmd, kvlist_t *args, kvlist_t *props);

/* return fd of a proc_id/proc_fd (proc_fd is the local fd from the
   proc's point of view, returned fd is a global plumb fd); proc_fd -1 means
   allocate the next free fd */
fd_t *fd_ref(char *proc_id, int proc_fd);

/* look up or allocate fd to a proc:fd; fd -1 means allocate the next free fd */
fd_t *fd_ref_(proc_virt_t *proc, int fd);

/* look up a process by id; returns whether the proc is real (0/1) or -1 if
   not found. When found, loads either pv or pr if they are not NULL */
int proc_lookup(const char *id, proc_virt_t **pv, proc_real_t **pr);

/* same, but calls plumb_error() for nonexisitng proc and returns the result
   casted to vp (useful for the parser) */
proc_virt_t *proc_lookup_safe(const char *id);

/* start all processes which are not yet started */
void proc_start_all(void);

/* stop processes and make some efforts to make sure they really die; if
   virt is zero, only real processes are stopped */
void proc_stop_all(int hard, int virt);


/* free the resources of an already stopped process */
void proc_free(proc_virt_t *p);

/* call proc_stop_all() and free up all resources */
void proc_free_all(void);

/* close fds 0, 1 and 2 for a specific or all process(es); if close_virt is
   zero, only real processes are closed */
void proc_close_stdio(proc_virt_t *p);
void proc_close_stdio_all(int close_virt);

/* return the number of real processes still in PS_RUNNING */
int proc_count_real_running(void);

/* process control */
void proc_stop(proc_virt_t *proc, int hard);   /* for real process terminate by signal or "kill -9" if hard is non-zero */
void proc_start(proc_virt_t *proc);

/* pause or unpause (if pause == 0) a process */
void proc_pause(proc_virt_t *proc, int pause);

/* check if process property kill-on-eof is set and all sinks are closed; if
   so, kill the process */
void proc_kill_on_eof(proc_virt_t *proc);

/* tells whether fd is a source or a sink */
#define is_source(fd)   (((fd)->pipe != NULL) && ((fd)->pipe->from == (fd)))
#define is_sink(fd)     (((fd)->pipe != NULL) && ((fd)->pipe->to   == (fd)))

#endif
