/*  plumb - the pipe shell - plumb script variables
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "genht/htsp.h"
#include "genht/hash.h"
#include "parse_common.h"
#include "plumb.h"
#include "plumb_lex.h"
#include "var.h"

static htsp_t *vars = NULL;
static void var_init(transaction_t *tx)
{
	if (vars == NULL)
		vars = htsp_alloc(strhash, strkeyeq);
	if ((tx != NULL) && (tx->vars == NULL))
		tx->vars = htsp_alloc(strhash, strkeyeq);
}

static void var_free_entry(htsp_entry_t *e)
{
	if (e->value != NULL)
		free(e->value);
	if (e->key != NULL)
		free(e->key);
}

static htsp_t *var_free_all(htsp_t *ht)
{
	htsp_entry_t *e;

	for (e = htsp_first(ht); e; e = htsp_next(ht, e))
		var_free_entry(e);
	htsp_free(ht);
	return NULL;
}

void var_uninit(transaction_t *tx)
{
	if (tx != NULL) {
		if (tx->vars != NULL)
			tx->vars = var_free_all(tx->vars);
	}
	else if (vars != NULL)
		vars = var_free_all(vars);
}

static var_t *var_value(const char *value, int vlen)
{
	var_t *v;
	v = malloc(vlen + sizeof(var_t));
	v->flags = 0;
	v->vlen = vlen;
	memcpy(v->value, value, vlen);
	return v;
}

static void var_set_v(htsp_t *ht, const char *name, var_t *v)
{
	htsp_entry_t *e;

	e = htsp_popentry(ht, (char *)name);
	if (e != NULL)
		var_free_entry(e);
	htsp_set(ht, strclone(name), v);
}


static var_t *var_set_(htsp_t *ht, const char *name, const char *value, int vlen)
{
	var_t *v;
	v = var_value(value, vlen);
	var_set_v(ht, name, v);
	return v;
}

var_t *var_set(transaction_t *tx, const char *name, const char *value)
{
	var_init(tx);
	if (tx == NULL)
		return var_set_(vars, name, value, strlen(value)+1);
	return var_set_((htsp_t *)tx->vars, name, value, strlen(value)+1);
}

var_t *var_get_(transaction_t *tx, const char *name)
{
	var_t *v;
	var_init(tx);

	if (tx != NULL)
		v = htsp_get((htsp_t *)tx->vars, (char *)name);
	else
		v = NULL;

	if (v == NULL)
		v = htsp_get(vars, (char *)name);

	return v;
}


const char *var_get(transaction_t *tx, const char *name)
{
	var_t *v;
	v = var_get_(tx, name);
	if (v == NULL)
		return NULL;
	return v->value;
}

void var_del(transaction_t *tx, const char *name)
{
	var_init(tx);

	if (tx != NULL)
		var_free_entry(htsp_pop((htsp_t *)tx->vars, (char *)name));
	
	var_free_entry(htsp_pop(vars, (char *)name));
}

void var_flush(transaction_t *tx, int destroy)
{
	htsp_entry_t *e;

	if (tx->vars == NULL)
		return;

	for (e = htsp_first(tx->vars); e; e = htsp_next(tx->vars, e)) {
		var_t *v2, *v = e->value;
		if (!(v->flags & VAR_NOFLUSH)) {
			v2 = var_set_(vars, e->key, v->value, v->vlen);
			v2->flags = v->flags;
		}
	}

	if (destroy) {
		var_uninit(tx);
		tx->vars = NULL;
	}
}

int yy_var_subst(transaction_t *tx, char *name, int prot, void *scanner)
{
	const char *value;

	name++;
	if (prot) {
		int len;
		name++;
		len = strlen(name);
		name[len-1] = '\0';
	}

	value = var_get(tx, name);
	if (value != NULL)
		yy_scan_string(value, scanner);
	else
		plumb_error_("Unknown/uninitialized variable %s\n", name);
	return 0;
}

