/*  plumb - the pipe shell - expose plumb env fds
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libporty/net/network.h>
#include <libporty/net_async/astdio.h>
#include "process.h"
#include "pipe.h"
#include "plumb.h"
#include "virt.h"

/* env is a dummy: syntactically virtual, but since it has external fd
   connections it must be a real process.  To overcome this problem, 
   a mirror real process _env is created. _env steals each env:X connection,
   connecting _env:X to the original target. _env is initialized by vp_env_init()

   Since stdio (fd 1 and 2) is handled specially on windows, we need to use
   libporty (astdio) for reading/writing it.

   This trick is required to keep things simple; an alternative
   would be to just look at the highest fd allocated by user connection to env
   and then allocate higher FDs and connect them to bare fds. This would
   require special read/write code while using a fake real process allows
   us to use the normal real process I/O.
*/

static P_net_socket env_stdin = P_NET_INVALID_SOCKET, env_stdout = P_NET_INVALID_SOCKET;
static proc_real_t *_env = NULL;

static void stdio_init()
{
	if (env_stdin == P_NET_INVALID_SOCKET) {
		if (P_net_async_stdio(&env_stdin, &env_stdout) != 0)
			plumb_error_("can't initialize async stdio\n");
	}
}

static virt_ctrl_t env_newfd(proc_virt_t *p, int n)
{
	int fd_num;

	if (is_source(&p->h.fds[n])) {
		if (n == 0) {
			stdio_init();
			fd_num = env_stdin;
		}
		else
			fd_num = n;
	}
	else if (is_sink(&p->h.fds[n])) {
		if (n == 1) {
			stdio_init();
			fd_num = env_stdout;
		}
		else
			fd_num = n;
	}
	vp_shadow_rebind(p, n, _env, fd_num);
	return VC_OK;
}

static virt_ctrl_t env_start(proc_virt_t *p)
{
	int n;

	/* bind FDs according to their direction in env */
	for(n = 0; n < p->h.num_fds; n++)
		env_newfd(p, n);

	return VC_OK;
}

static virt_ctrl_t env_panic(proc_virt_t *proc, int local_fd)
{
	plumb_error_("Direct fd call to env\n");
	return VC_QUIT;
}

static virt_ctrl_t env_panic_i(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	return env_panic(proc, local_fd);
}

void _env_stop(proc_real_t *p, int hard)
{
	/* for now just mark the process dead and let plumb infra close all fds,
	   this way stderr can be written during the shutdown procedure */
	p->h.state = PS_STOPPED;
}

/* init the visible env VP */
int vp_env_init(proc_virt_t *p, kvlist_t *args)
{
	/* make sure the user can not initialize new envs */
	if (_env != NULL)
		plumb_error_("virtual process of env can not be manually created\n");

	p->newfd = env_newfd;

	_env = vp_shadow_create(p, strclone("_env"));
	_env->stop = _env_stop;

	/* allocate a process descriptor */
	p->input = env_panic_i;
	p->sink_eof = env_panic;
	p->source_eof = env_panic;
	p->timer = NULL;
	p->start = env_start;
	/* link them together */
	p->data = &_env;


	vp_shadow_create_finish(_env, 1);
	return 0;
}
