/*  plumb - the pipe shell - open files on the host file system
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libporty/net/network.h>
#include <libporty/net_async/astdio.h>
#include "process.h"
#include "pipe.h"
#include "plumb.h"
#include "virt.h"
#include "_default.h"

static void open_shadow_stop(proc_real_t *p, int hard)
{
	proc_stop(p->shadow_of, hard);
}

static virt_ctrl_t open_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	if (local_fd != vstdctrl)
		return vp_input_shadowed_panic(proc, local_fd, data, len);
	return VC_OK;
}

static virt_ctrl_t open_start(proc_virt_t *p)
{
	int real_fd;
	int oflags;
	proc_real_t *shadow;
	char *sflags;

	shadow = vp_shadow_create(p, NULL);
	shadow->h.kill_on_eof = 1;
	shadow->stop = open_shadow_stop;

	/* check for invalid pipe directions */
	if (vp_is_sink(p, vstdout))
		plumb_error_("stdout of %s=[open] is a sink\n");
	if (vp_is_source(p, vstdin))
		plumb_error_("stdin of %s=[open] is a source\n");
	if (vp_is_source(p, vstdctrl))
		plumb_error_("stdctrl of %s=[open] is a source\n");
	if (vp_is_sink(p, vstderr))
		plumb_error_("stderr of %s=[open] is a sink\n");

	/* at this point if an fd is bound, it is valid and usable */
	if ((vp_is_bound(p, vstdin)) && (vp_is_bound(p, vstdout))) {
		oflags = O_RDWR;
		sflags = "reading and writing";
	}
	else if (vp_is_bound(p, vstdout)) {
		oflags = O_RDONLY;
		sflags = "reading";
	}
	else {
		oflags = O_WRONLY | O_CREAT;
		sflags = "writing";
	}

	real_fd = open(p->data, oflags, 0640);
	if (real_fd < 0) {
		/* if stderr is bound, report error there instead of bringing lumb down */
		if (vp_is_bound(p, vstderr)) {
			char *msg;
			int len;
			msg = malloc(strlen(p->data) + 32);
			len = sprintf(msg, "can not open %s for %s\n", (char *)p->data, sflags);
			vp_write(p, vstderr, msg, len);
			free(msg);
			return VC_QUIT;
		}
		plumb_error_("%s=[open]: can not open %s for %s\n", p->h.id, p->data, sflags);
	}

	/* bind stdin and stdout */
	if (vp_is_bound(p, vstdout))
		vp_shadow_rebind(p, vstdout, shadow, real_fd);
	if (vp_is_bound(p, vstdin))
		vp_shadow_rebind(p, vstdin, shadow, real_fd);


	vp_shadow_create_finish(shadow, 0);

	free(p->data);

	return VC_OK;
}


/* init the visible open VP */
int vp_open_init(proc_virt_t *p, kvlist_t *args)
{
	/* allocate a process descriptor */
	p->input = open_input;
	p->sink_eof = vp_eof_shadowed_panic;
	p->source_eof = vp_eof_shadowed_panic;
	p->timer = NULL;
	p->start = open_start;
	p->data = strclone(args->head->key);

	return 0;
}
