/*  plumb - the pipe shell - regex match and/or replace
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "process.h"
#include "fd.h"
#include "genregex/regex_be.h"
#include "virt.h"

typedef enum {
	PASS_INVALID,
	PASS_MATCH,
	PASS_MISMATCH,
	PASS_ALL,
} pass_t;

typedef struct {
	pass_t pass;
	int global;
	char *subst;
	int substlen;
	re_be_t *re;
} regex_t;

static virt_ctrl_t regex_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	regex_t *r = proc->data;
	int match;

	match = re_be_exec(r->re, data, len);

	if (r->subst == NULL) { /* match-only mode */
		if ((match && (r->pass == PASS_MATCH)) || (!match && (r->pass == PASS_MISMATCH)))
			vp_write(proc, 1, data, len);
		return VC_OK;
	}

	/* subst mode */
	if ((!match) && (r->pass == PASS_MATCH))
		return VC_OK;

	if (match) { /* subst on match */
		char *buff;
		int bufflen;

		bufflen = re_be_subst(r->re, &buff, data, len, r->subst, r->substlen, !r->global);

		vp_write(proc, 1, buff, bufflen);
	}
	else /* print on mismatch */
		vp_write(proc, 1, data, len);


	return VC_OK;
}

void regex_free(proc_virt_t *proc)
{
	regex_t *r = proc->data;

	re_be_free(r->re);
	if (r->subst != NULL)
		free(r->subst);
	free(r);
}



/* compile a regex with special exceptions for $ */
static re_be_t *regex_compile(const char *pat)
{
	int l;
	char *cpy;
	re_be_t *ret;

	l = strlen(pat);

	/* if pattern doesn't end in end-of-line, just compile it */
	if ((l <= 1) || (pat[l-1] != '$'))
		return re_be_comp(pat);

	/* double end-of-line means binary end-of-record - pass on with a single $ */
	if ((l > 2) && (pat[l-2] == '$')) {
		cpy = strclone(pat);
		cpy[l-1] = '\0';
		ret = re_be_comp(cpy);
		free(cpy);
		return ret;
	}

	/* single end-of-line means text: check for newline */
	cpy = malloc(l+8);
	memcpy(cpy, pat, l-1);
	strcpy(cpy+l-1, "[\\r\\n]+$");
	ret = re_be_comp(cpy);
	free(cpy);
	return ret;
}

int vp_regex_init(proc_virt_t *p, kvlist_t *args)
{
	regex_t *m;
	char *pattern = NULL, *spass = NULL, *subst = NULL, *global = NULL;
	pass_t pass = PASS_INVALID;
	vp_args_t vargs[] = {
		{"pattern", &pattern},
		{"pass",    &spass},
		{"subst",   &subst},
		{"global",  &global},
		{NULL, NULL}
	};

	vp_stdfds_set(p, VP_REQ_QUIT_ON_UNBIND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_UNBOUND, VP_REQ_UNBOUND, VP_REQ_UNBOUND, VP_REQ_UNBOUND);
	vp_args(args, vargs);

	if (pattern == NULL)
		plumb_error_("regex %s without pattern\n", p->h.id);

	if (spass != NULL) {
		if ((spass[1] == 'i') || (spass[1] == 'I'))
			pass = PASS_MISMATCH;
		else if ((spass[0] == 'a') || (spass[0] == 'A'))
			pass = PASS_ALL;
		else if ((spass[1] == 'a') || (spass[1] == 'A'))
			pass = PASS_MATCH;
		else
			plumb_error_("regex %s: invalid pass mode %s\n", p->h.id, spass);
	}
	else {
		/* default pass mode is match for matching and all for subst*/
		pass = subst == NULL ? PASS_MATCH : PASS_ALL;
	}
	if ((global != NULL) && (subst == NULL))
		plumb_error_("regex %s has global= but does not have subst=\n", p->h.id);

	if ((pass == PASS_MISMATCH) && (subst != NULL))
		plumb_error_("regex %s tries to do substitution and pass on mismatches\n", p->h.id);

	if ((pass == PASS_ALL) && (subst == NULL))
		plumb_error_("regex %s tries to match and pass all, regardless of matching\n", p->h.id);

	m = malloc(sizeof(regex_t));
	m->pass     = pass;
	m->re       = regex_compile(pattern);
	m->global   = vp_istrue(global);
	m->subst    = strclone(subst);
	m->substlen = subst == NULL ? 0 : strlen(subst);

	p->input = regex_input;
	p->data  = m;
	p->free  = regex_free;

	return 0;
}
