/*  plumb - the pipe shell - create network sockets
    Copyright (C) 2020  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libporty/net/network.h>
#include <libporty/net/tcp4.h>
#include "process.h"
#include "fd.h"
#include "libporty/util/buf.h"
#include "virt.h"
#include "pipe.h"
#include "timer.h"
#include "_default.h"
#include "config.h"

typedef enum {
	SP_INVALID = 0,
	SP_TCP4_CONNECT,
	SP_TCP4_LISTEN
} socket_proto_t;

typedef enum {
	SA_SINGLE = 0,            /* one-shot: connect once and quit if the connection is broken */
	SA_IDLE,                  /* persistent, but no automatic connection */
	SA_RECONN                 /* persistent, automatic connection and reconnection after some time */
} socket_ambition_t;

#define FD_SOCKFD 100
#define FD_RECVFD 101

typedef struct {
	socket_proto_t proto;
	char *host, *lochost;
	int port, locport;
	fd_t *netfd; /* the socket's real fd */
	fd_t *recvfd; /* the fd tied to [socket] where the virtual process receives the reads */
	P_net_socket s;
	socket_ambition_t ambition;
	int reconn;
} socket_t;


static int vp_socket_set_proto(proc_virt_t *proc, const char *proto)
{
	socket_t *sk = proc->data;

	if (strcmp(proto, "tcp4-connect") == 0) sk->proto = SP_TCP4_CONNECT;
	else if (strcmp(proto, "tcp4-listen") == 0) sk->proto = SP_TCP4_LISTEN;
	else return -1;

	return 0;
}

static void socket_detect_conn(P_net_socket sockfd, fd_t *fd)
{
	proc_virt_t *proc = fd->proc.v;
	vp_event_printf(proc, "socket", 128, "connection established");
	fd_del_watch(fd, P_POLLOUT);
	fd->override_do_write = NULL;
}

static void vp_socket_try_connect(proc_virt_t *proc, int emit_error)
{
	socket_t *sk = proc->data;

	if ((sk->host == NULL) || (sk->port == 0) || (sk->proto == SP_INVALID)) {
		if (emit_error)
			vp_event_printf(proc, "socket", 128, "can not connect: missing host, port or proto");
		return;
	}

	switch(sk->proto) {
		case SP_INVALID: break;
		case SP_TCP4_CONNECT:
			vp_event_printf(proc, "socket", 128, "connection attempt");
			sk->netfd->override_do_write = socket_detect_conn;
			fd_add_watch(sk->netfd, P_POLLOUT);
			if (P_tcp4_connect(&sk->s, sk->host, sk->port, sk->lochost, sk->locport, P_net_nonblock_full) != 0) {
				vp_event_printf(proc, "socket", 128, "connection failed");
				fd_del_watch(sk->netfd, P_POLLOUT);
			}
			else {
				/* undo the effect of a previous eof */
				sk->netfd->eof = 0;
				sk->netfd->closing = 0;
				sk->recvfd->eof = 0;
				sk->recvfd->closing = 0;
				
				/* set up polling on netfd and the pipe so that data is received on recvfd */
				fd_set_socket(sk->netfd, sk->s);
				fd_add_watch(sk->netfd, P_POLLIN);
				pipe_register(sk->netfd, sk->recvfd);
			}
			break;
		case SP_TCP4_LISTEN:
			fprintf(stderr, "TODO: no listen yet\n");
			abort();
	}
}


static virt_ctrl_t socket_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	socket_t *sk = proc->data;

	fprintf(stderr, "socket %s: input on %d: %d\n", proc->h.id, local_fd, len);

	switch(local_fd) {
		case FD_RECVFD: vp_write(proc, vstdout, data, len); break;
		case 0:
			if (sk->s != P_NET_INVALID_SOCKET)
				fd_write_sockfd(sk->netfd, sk->s, data, len);
			else
				vp_event_printf(proc, "socket", 128, "send ignored: no connection");
			break;
	}

	return VC_OK;
}

static void socket_free(proc_virt_t *proc)
{
	socket_t *sk = proc->data;
	free(sk->host);
	free(sk);
}

static virt_ctrl_t socket_close(proc_virt_t *proc)
{
	socket_t *sk = proc->data;

	if (sk->s != P_NET_INVALID_SOCKET) {
		P_net_close(sk->s);
		fd_del_watch(sk->netfd, P_POLLOUT);
		fd_clear_buff(sk->netfd);
		sk->s = P_NET_INVALID_SOCKET;
		vp_event_printf(proc, "socket", 128, "connection lost");
		if (sk->ambition == SA_RECONN)
			timer_add_delta(proc, sk->reconn, NULL);
	}

	if (sk->ambition == SA_SINGLE)
		return VC_QUIT;

	return VC_OK;
}

/* eof on sink (input): may need to buffer it */
virt_ctrl_t socket_sink_eof(proc_virt_t *proc, int local_fd)
{
	if (local_fd == FD_RECVFD)
		return socket_close(proc);

	socket_close(proc);
	return VC_QUIT;
}

static virt_ctrl_t socket_newfd(proc_virt_t *p, int n)
{
	switch(n) {
		case FD_SOCKFD:
		case FD_RECVFD:
			break;
		default:
			vp_default_newfd(p, n);
	}
	return VC_OK;
}

static virt_ctrl_t socket_timer(proc_virt_t *proc, void *data)
{
	socket_t *sk = proc->data;

	if ((sk->ambition != SA_IDLE) && (sk->s == P_NET_INVALID_SOCKET))
		vp_socket_try_connect(proc, 0);
	return VC_OK;
}


static virt_ctrl_t socket_start(proc_virt_t *proc)
{
	socket_t *sk = proc->data;

	if (sk->ambition != SA_IDLE)
		vp_socket_try_connect(proc, 0);

	return VC_OK;
}

int vp_socket_init(proc_virt_t *p, kvlist_t *args)
{
	socket_t *sk;
	char *sproto = NULL, *shost = NULL, *sport = NULL, *sreconn = NULL;
	vp_args_t vargs[] = {
		{"proto", &sproto},
		{"host",  &shost},
		{"port",  &sport},
		{"reconn",&sreconn},
		{NULL, NULL}
	};

	vp_args(args, vargs);

	sk = calloc(1, sizeof(socket_t));
	p->data = sk;
	sk->s = P_NET_INVALID_SOCKET;

	if (sproto != NULL)
		vp_socket_set_proto(p, sproto);
	if (shost != NULL)
		sk->host = strclone(shost);
	if (sport != NULL)
		sk->port = vp_arg_atoi("[socket]", "port", sport);
	if (sreconn != NULL) {
		if (strcmp(sreconn, "single") == 0) {
			sk->ambition = SA_SINGLE;
		}
		else if (strcmp(sreconn, "idle") == 0) {
			sk->ambition = SA_IDLE;
		}
		else {
			sk->reconn = vp_arg_atoi("[socket]", "reconn", sreconn);
			if (sk->reconn > 0)
				sk->ambition = SA_RECONN;
			else if (sk->reconn < 0)
				sk->ambition = SA_IDLE;
		}
	}

	p->input        = socket_input;
	p->free         = socket_free;
	p->sink_eof     = socket_sink_eof;
	p->newfd        = socket_newfd;
	p->start        = socket_start;
	p->timer        = socket_timer;


	vp_stdfds_set(p, VP_REQ_QUIT_ON_UNBIND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_UNBOUND, VP_REQ_UNBOUND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_UNBOUND);

	sk->netfd = fd_alloc();
	sk->netfd->proc.v = p;
	sk->netfd->fd = FD_SOCKFD;
	sk->recvfd = fd_alloc();
	sk->recvfd->proc.v = p;
	sk->recvfd->fd = FD_RECVFD;

	return 0;
}
